use std::sync::Arc;
use tokio::sync::RwLock;
use hyper::upgrade::Upgraded;
use hyper_util::rt::TokioIo;
use tokio_tungstenite::WebSocketStream;
use futures::StreamExt;

use crate::{
    protocol,
    util::pack::unpack_u16,
    HttpServiceArgs,
};

pub async fn handle_ws(ws:WebSocketStream<TokioIo<Upgraded>>, args:HttpServiceArgs) -> Result<(),()>
// Handle websocket connection.
//
{
    use tokio_tungstenite::tungstenite::protocol::Message;
    use protocol::{QRPacket, QRPacketData, code::*, packet::*};

    let conn_id :u32;

    let (sink, mut stream) = ws.split();

    let bus_ds = args.bus.mailbox(1).unwrap_or(1);

    // Perform connection handshake with data system.
    //  - Provide system with connection/bus pairing.
    //  - Acquire connection id.
    //
    args.bus.send(bus_ds, QRPacket::new(0, QRPacketData::QConn(LocalPacketConnect {
        bus_id:args.bus.id(),
        stream:Arc::new(RwLock::new(sink)),
    })))?;
    match args.bus.receive_wait() {
        Some(resp) => {
            let qr = &resp.data;
            match qr.data {
                QRPacketData::RConn => { conn_id = qr.id; }
                _ => { return Err(()); }
            }
        }
        None => { return Err(()); }
    }

    // Decode client requests from websocket,
    // pass requests to data system,
    // and return responses to client.
    while match stream.next().await {
        Some(msg) => match msg {
            Ok(msg) => match msg {
                Message::Binary(data) => {
                    let mut index :usize = 0;
                    let code: u16 = unpack_u16(&data, &mut index);

                    println!("MESSAGE {:x}", code);

                    match code {
                        CODE_HELLO => {
                            args.bus.send(
                                bus_ds, QRPacket::new(conn_id, QRPacketData::QHello)
                            ).ok();
                        }

                        CODE_REGISTER => match PacketRegister::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QRegister(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_AUTH => match PacketAuth::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QAuth(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_AUTH_RESUME => match PacketAuthResume::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QAuthResume(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_AUTH_REVOKE => {
                            args.bus.send(
                                bus_ds, QRPacket::new(conn_id,
                                    QRPacketData::QAuthRevoke)
                            ).ok();
                        }

                        CODE_SESSION_LIST => match PacketSessionList::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QSessionList(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        /*CODE_SESSION_JOIN => match PacketSessionCreate::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QSessionCreate(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }*/

                        CODE_SESSION_VIEW => match PacketSessionView::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QSessionView(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_SESSION_LEAVE => {
                            args.bus.send(
                                bus_ds,
                                QRPacket::new(conn_id, QRPacketData::QSessionLeave)
                            ).ok();
                        }

                        CODE_SESSION_RETIRE => match PacketSessionRetire::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QSessionRetire(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_GAME_STATE => match PacketGameState::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QGameState(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_GAME_MESSAGE => match PacketGameMessage::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::GameMessage(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_CHALLENGE => match PacketChallenge::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QChallenge(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_CHALLENGE_ANSWER => match PacketChallengeAnswer::decode(&data, &mut index) {
                            Ok(packet) => {
                                args.bus.send(
                                    bus_ds,
                                    QRPacket::new(conn_id, QRPacketData::QChallengeAnswer(packet))
                                ).ok();
                            }
                            Err(_) => { println!("error: packet decode failed."); }
                        }

                        CODE_CHALLENGE_LIST => {
                            args.bus.send(
                                bus_ds,
                                QRPacket::new(conn_id, QRPacketData::QChallengeList)
                            ).ok();
                        }

                        CODE_USER_LIST => {
                            args.bus.send(
                                bus_ds,
                                QRPacket::new(conn_id, QRPacketData::QUserList)
                            ).ok();
                        }

                        _ => { }
                    }
                    true
                }
                
                Message::Close(_) => { true }

                _ => {
                    println!("notice: received unexpected websocket data.");
                    true
                }
            }
            Err(e) => {
                println!("WSS error: {}", e.to_string());
                false
            },
        }
        None => false,
    } { }

    args.bus.send(bus_ds, QRPacket::new(conn_id, QRPacketData::QDisconn)).ok();
    Ok(())
}
